#pragma once
#include <string>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <vector>

class MyException : public std::exception
{
	const char* exception;
public:
	MyException(const char* exception_p)
	{
		this->exception = exception_p;
	}
	const char* what() const { return this->exception; }
};

void createDataStream(const std::string&);//!!! Строки, как и все объекты, передаются по ссылке на контанту

void createDataIterator(const std::string&);//!!! Строки, как и все объекты, передаются по ссылке на контанту

void ModifyContainerTransform(const std::string&, const std::string&);//!!! Строки, как и все объекты, передаются по ссылке на контанту

template < typename T >
void ModifyContainer(T&);



void createDataStream(const std::string& filename_p) //!!! Строки, как и все объекты, передаются по ссылке на контанту
{
	std::ofstream fout(filename_p);
	std::srand(std::time(nullptr));
	for (int i = 0; i < 100; i++)
	{
		fout << std::rand() % 101 - 50 << " ";
	}

}


void createDataIterator(const std::string& filename_p)//!!! Строки, как и все объекты, передаются по ссылке на контанту
{
	std::srand(std::time(nullptr));
	std::vector<int> fcontainer;
	fcontainer.reserve(100);
	std::generate_n(std::back_inserter(fcontainer), 100, [] { return std::rand() % 101 - 50; });
	std::ofstream fout(filename_p);
	std::copy(fcontainer.begin(), fcontainer.end(), std::ostream_iterator<int>(fout, " "));
}


template<typename T>
T loadDataStream(const std::string& filename_p)//!!! Строки, как и все объекты, передаются по ссылке на контанту
{
	T outputContainer;
	std::ifstream fin(filename_p);
	if (fin.fail())
		throw MyException("Could not find file");
	typename T::value_type x;
	while (fin >> x)
		outputContainer.push_back(x);
	return outputContainer;
}


template<typename T>
T loadDataIterator(const std::string& filename_p)//!!! Строки, как и все объекты, передаются по ссылке на контанту
{
	T outputContainer;
	std::ifstream fin(filename_p);
	if (fin.fail())
		throw MyException("Could not find file");

	std::copy(std::istream_iterator<typename T::value_type>(fin), std::istream_iterator<typename T::value_type>(), std::back_inserter(outputContainer));
	return outputContainer;

}

template < typename T >
void ModifyContainer(T& container_p)
{
	if (container_p.size() == 0)
		throw MyException("Can not modify empty container");

	auto last_number = *(--container_p.end());
	auto max_element = *std::max_element(container_p.begin(), container_p.end());
	auto number_to_add = sqrt(last_number * max_element);

	typename T::value_type sum = 0;
	for (auto it = container_p.begin(); it != container_p.end(); ++it)
	{
		*it += number_to_add;
		sum += abs(*it);
	}
	container_p.push_back(sum);
	container_p.push_back(sum / container_p.size());
}

template < typename T >
void ModifyContainerIterators(T begin, T end)
{
	if (begin == end)
		throw MyException("Can not modify empty container");

	auto last_number = *(--end);
	end++;
	auto max_element = *std::max_element(begin, end);
	auto number_to_add = sqrt(last_number * max_element);

	for (auto it = begin; it != end; ++it)
	{
		*it += number_to_add;
	}
}

template < typename T >
void ModifyContainerForEach(T& container_p)
{
	if (container_p.size() == 0)
		throw MyException("Can not modify empty container");
	auto last_number = *(--container_p.end());
	auto max_element = *std::max_element(container_p.begin(), container_p.end());
	auto number_to_add = sqrt(last_number * max_element);

	std::for_each(container_p.begin(), container_p.end(), [number_to_add](typename T::value_type& el) { return el += number_to_add; });
}


void ModifyContainerTransform(const std::string& filenameIn_p, const std::string& filenameOut_p)
{
	std::ifstream fin(filenameIn_p);
	if (fin.fail())
		throw MyException("Could not find file");

	std::vector<int> tempContainer;
	std::copy(std::istream_iterator<int>(fin), std::istream_iterator<int>(), std::back_inserter(tempContainer));
	if (tempContainer.size() == 0)
		throw MyException("Can not modify empty container");
	auto last_number = *(--tempContainer.end());
	auto max_element = *std::max_element(tempContainer.begin(), tempContainer.end());
	auto number_to_add = sqrt(last_number * max_element);
	std::ofstream fout(filenameOut_p);
	fin.clear();
	fin.seekg(0);

	std::transform(std::istream_iterator<int>(fin), std::istream_iterator<int>(), std::ostream_iterator<int>(fout, " "),
		[number_to_add](int el) { return el + number_to_add; });
}


template < typename T >
void outputResultStream(const T& container_p, const std::string file_name)
{
	std::ofstream fout(file_name);
	for (auto it = container_p.cbegin(); it != container_p.cend(); ++it)
	{
		fout << *it << " ";
	}
}

template < typename T >
void outputResultCopy(const T& container_p, const std::string& file_name)
{
	std::ofstream fout(file_name);

	std::copy(container_p.begin(), container_p.end(), std::ostream_iterator<int>(fout, " "));
}

template < typename T >
void printContainer(const T& container_p)
{
	for (auto it = container_p.cbegin(); it != container_p.cend(); ++it)
		std::cout << *it << " ";
}